#include <math.h>
#include <stdlib.h> // For rand()

#include <iostream> // For std::cerr

#include <GL/gl.h>

#include <qimage.h> // To load the texture image.
#include <qgl.h> // To convert the texture image with convertToGLFormat()..

#include "dynamicCanvas.h"

using namespace std;

DynamicCanvas::DynamicCanvas():
  warped(true),
  frequencyBase_(1.0),
  paperTransparency(0.3),
  D(1.0),
  posX(0),
  posY(0),
  alpha(0),
  beta(0),
  gamma(0),
  deltaAlpha(0),
  deltaBeta(0),
  deltaGamma(0),
  previousCameraPosition(0,0,0),
  omega0(0),
  nbOctaves_(4),
  textureSize_(512),
  nbOffset(128),
  zoom(1.0)
{
  // Resize the random offset array.
  randomOffsetX.resize(nbOffset);
  randomOffsetY.resize(nbOffset);

  // Fill the array with random values in [0;1].
  for(unsigned int i=0;i<nbOffset;i++){
    randomOffsetX[i] = static_cast<float>(rand())/RAND_MAX;
    randomOffsetY[i] = static_cast<float>(rand())/RAND_MAX;
  }
}

DynamicCanvas::~DynamicCanvas()
{  
  // NEED TO BE CHECKED : MAYBE REQUIRE A DESTROYGL FUNCTION. MAYBE USELESS.
  // If the display list still exists, delete it.
  if (glIsList(paperGridID) == GL_TRUE)
    glDeleteLists(paperGridID,1);
}

bool DynamicCanvas::loadPaperTexture(const char* name)
{
  QImage image(name);

  // Check if image has been loaded.
  if( image.isNull() ) {
    std::cerr<<"DynamicCanvas::loadPaperTexture():\n"
	     <<"Unable to load image \"" << name << "\"."
	     <<std::endl;
    return false;
  }

  // //Check image format.
  // if( image.depth() != 32 ) {
    // std::cerr<<"DynamicCanvas::loadPaperTexture():\n"
	     // <<"Image not in RGBA mode (depth="<<image.depth()<<")"
	     // <<std::endl;
    // return false;
  // }

  QImage glImage = QGLWidget::convertToGLFormat(image);  // flipped 32bit RGBA

  // Delete existing texture.
  if (glIsTexture(paperTextureID)){
    glDeleteTextures(1,&paperTextureID);
  }

  // Create new texture.
  glPixelStorei(GL_UNPACK_ALIGNMENT,1) ; 
  glGenTextures(1,&paperTextureID);
  glBindTexture(GL_TEXTURE_2D, paperTextureID);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexImage2D(GL_TEXTURE_2D,0,GL_RGB,image.width(),image.height(),0,GL_BGRA,GL_UNSIGNED_BYTE,glImage.bits());

  // Generate the multi-scale texture.
  buildMultiScaleTexture();

  return true;
}


void DynamicCanvas::initGL(const unsigned int width,
			   const unsigned int height,
			   const qglviewer::Camera& camera)
{
  /////////
  // Multi-scale texture initialization
  ////////  
  
  /*
    Remark on the maximum composite texture size
    --------------------------------------------
  
    Consider that example: with an input paper texture of size 64x64, and 5 octaves,
    you need to composite with a size of 64*2^{5-1} = 1024x1024 to
    be sure to avoid any subsampling...
  */
  
  // Optimal composition size search. We search for the biggest
  // texture size that fits in the current window.
  while ((textureSize_ > width) || (textureSize_ > height))
    textureSize_ /= 2;

  // Initialize texture to be loaded with the octave composition (see buildMultiScaleTexture() )
  GLubyte *composite = new GLubyte[textureSize_*textureSize_*4];

  // Texture unit initialization for the multi-scale texture:
  // smooth interpolation (GL_LINEAR) + tiled texture (GL_REPEAT).
  glGenTextures(1,&msTextureID);
  glBindTexture(GL_TEXTURE_2D,msTextureID);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S,     GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T,     GL_REPEAT);

  // We reclaim the texture memory. No need to send meaningful data,
  // they will be overwritten later.
  glTexImage2D(GL_TEXTURE_2D,0,GL_RGB,textureSize_,textureSize_,0,GL_RGBA,GL_UNSIGNED_BYTE,composite);
	
  delete[] composite;

  /////////
  // Initialize the display list.
  /////////
  
  buildDisplayList();

  initialCameraOrientation_ = camera.orientation();
  previousCameraPosition = camera.position();
}


void DynamicCanvas::updateParameters(const qglviewer::Camera& camera)
{
  // Reference frame used for angle computation.
  static const qglviewer::Frame referenceFrame = *(camera.frame());


  
  ///////
  // Angle values : alpha, beta, gamma
  ///////

  
  // Let's define two useful vectors.
  // A is the Y (B is the -Z) of the camera, expressed in the initial camera orientation coordinate system.
  const vectorType A = initialCameraOrientation_.inverseRotate(camera.frame()->inverseTransformOf(vectorType(0,1,0)));
  const vectorType B = initialCameraOrientation_.inverseRotate(camera.frame()->inverseTransformOf(vectorType(0,0,-1)));

  
  // Computation of alpha.
  const float newAlpha = atan2f(sqrtf(A.x*A.x+A.z*A.z),A.y) * ((B.y>0) ? 1 : -1);

  // Computation of beta.
  float newBeta;
  if (fabsf(B.y) < fabsf(A.y))
    newBeta = atan2f(B.x,-B.z);
  else
    {
      if (B.y < 0)
	newBeta = atan2f(A.x,-A.z);
      else
	newBeta = atan2f(-A.x,A.z);
    }

  // Computation of the delta values while accounting for the -PI/2;+PI/2 jump.
  // A jump occurs if newValue*oldValue is negative enough.
  
  // deltaAlpha
  if (alpha*newAlpha < -1.23456f)
    deltaAlpha = (alpha>0) ? newAlpha+2.0f*M_PI-alpha : newAlpha-2.0f*M_PI-alpha;
  else
    deltaAlpha = newAlpha - alpha;
  

  // deltaBeta
  if (beta * newBeta < -1.23456f)
    deltaBeta = ( beta > 0.0) ? newBeta+2.0f*M_PI-beta : newBeta-2.0f*M_PI-beta;
  else 
    deltaBeta = newBeta - beta;
  

  // deltaGamma
  deltaGamma = sinf(alpha) * deltaBeta;

  
  // Update angle values.
  alpha  = newAlpha;
  beta   = newBeta;
  gamma += deltaGamma;


  
  ////////
  // Update the ratio and the field of view
  ////////

  const float newRatio = camera.aspectRatio();
  const bool ratioChanged = (ratio!=newRatio);
  ratio = newRatio;

  const float newFieldOfView = camera.fieldOfView();
  const bool fieldOfViewChanged = (fieldOfView!=newFieldOfView);
  fieldOfView = newFieldOfView;



  ///////
  // Update the zoom factor
  ///////
  
  const vectorType cameraPosition = camera.position();
  
  // deltaZ = deltaPosition . viewDirection (B in short)
  float deltaZ = (cameraPosition - previousCameraPosition) * camera.viewDirection();

  // See section 5.2.
  zoom *= (subjectiveDistance()-deltaZ) / subjectiveDistance();
  
  // Store the current position.
  previousCameraPosition = cameraPosition;

  sphereRadius = 1.0f/tanf(fieldOfView/2.0f);
  
  ////////
  // Update the display list if needed
  ////////

  if (ratioChanged || fieldOfViewChanged || (warped && (fabsf(deltaAlpha) > 1E-3)))
    buildDisplayList();

  ////////
  // Update the multi-scale texture if needed
  ////////

  if (fabsf(deltaZ) > 1E-3)
    buildMultiScaleTexture();
}


void DynamicCanvas::drawPaper()
{
  // Disable z-buffer, lighting and face culling.
  glDisable(GL_DEPTH_TEST);
  glDisable(GL_LIGHTING);
  glDisable(GL_CULL_FACE);

  // Set up the texture engine.
  glEnable(GL_TEXTURE_2D);
  glBindTexture(GL_TEXTURE_2D, msTextureID);
  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

  // Set up the blending function.
  // Formally, the blend equation is [paperTransparency*paper + (1-paperTransparency)*scene].
  // The glColor4f() cannot be used to specify the alpha value. So
  // we use the specially provided extention to do that :
  glEnable(GL_BLEND);
  glBlendEquation(GL_FUNC_ADD);
  glBlendFunc(GL_CONSTANT_ALPHA,GL_ONE_MINUS_CONSTANT_ALPHA);
  glBlendColor(1.0,1.0,1.0, paperTransparency); 
  
  // Set up the modelview matrix.
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity();

  // Set up the projection matrix.
  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  glLoadIdentity();
  glOrtho(-1.0, 1.0, -1.0, 1.0, -1.0f, 1.0f);

  // Save the texture matrix.
  glMatrixMode(GL_TEXTURE);
  glPushMatrix();

  // Set up the texture matrix.
  setTextureMatrix();

  // Draw the paper grid.
  glCallList(paperGridID);

  // Restore the matrices.
  glMatrixMode(GL_TEXTURE);    glPopMatrix();
  glMatrixMode(GL_PROJECTION); glPopMatrix();	
  glMatrixMode(GL_MODELVIEW);  glPopMatrix();
  
  glDisable(GL_TEXTURE_2D);
  glDisable(GL_BLEND);
  
  glEnable(GL_DEPTH_TEST);
  glEnable(GL_LIGHTING);
  glEnable(GL_CULL_FACE);
}


void DynamicCanvas::buildDisplayList()
{ 
  /////////
  // Constant definition
  /////////
  
  //! Number of lines/columns in the "paper" grid.
  const unsigned int lineNumber   = 20;
  const unsigned int columnNumber = 20;
    
  //! Dimension of a line/column. The grid goes from -1 to +1.
  const float lineSize   = 2.0f / lineNumber;
  const float columnSize = 2.0f / columnNumber;

    

  /////////
  // Display list creation
  /////////
    
  // If the list already exists, delete it.
  if(glIsList(paperGridID) == GL_TRUE)
    glDeleteLists(paperGridID,1);
  

  // Create a new display list.
  paperGridID = glGenLists(1);
    
  // Start creation of the display list.
  glNewList(paperGridID, GL_COMPILE);
  glBegin(GL_QUADS);

  for (unsigned int i=0; i<lineNumber; i++)
    {
      const float x = -1.0f + i*lineSize;
      for(unsigned int j=0; j<columnNumber; j++)
	{
	  const float y = -1.0f + j*columnSize;

	  // Creation of the 4 corners of the (i,j) cell.
	  textureCoordinate(x,y);
	  glVertex3f(x,y,0.0);

	  textureCoordinate(x+lineSize,y);
	  glVertex3f(x+lineSize,y,0.0);

	  textureCoordinate(x+lineSize,y+columnSize);
	  glVertex3f(x+lineSize,y+columnSize,0.0);

	  textureCoordinate(x,y+columnSize);
	  glVertex3f(x,y+columnSize,0.0);
	  // Cell creation done.
	}
    }

  glEnd();
  glEndList();
  // The display list is created.
}


void DynamicCanvas::buildMultiScaleTexture()
{
  // The zoom factor is designed to be folded into range 1..2 : if it
  // becomes greater than 2 or lower than 1 we have to shift the octaves
  // and their random translation index. See section 5.2.

  // Shift octave: apparition of high frequency = going forward.
  while(zoom >= 2.0)
    {
      // Decrease omage0 but keep it cyclic.
      omega0 = (omega0+nbOffset-1) % nbOffset;

      // Update zoom and position. See section 5.2
      zoom *= 0.5;
      posX *= 0.5;
      posY *= 0.5;
    }

  // Shift octave: apparition of low frequency = going backward.
  while(zoom < 1.0)
    {
      // Increase omage0 but keep it cyclic.
      omega0 = (omega0+1) % nbOffset;

      // Update zoom and position. See section 5.2
      zoom *= 2.0;
      posX *= 2.0;
      posY *= 2.0;
    }
  
  // Reset the buffer.
  glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
  glClear(GL_COLOR_BUFFER_BIT);

  // Disable some unused features.
  glDisable(GL_DEPTH_TEST);
  glDisable(GL_COLOR_MATERIAL);
  glDisable(GL_LIGHTING);

  // The blend color is needed to have a constant value during the
  // blending of the octaves.
  // glBlendColor(1.0f, 1.0f, 1.0f, 1.0f);
  glEnable(GL_BLEND);
  glBlendFunc(GL_CONSTANT_ALPHA, GL_ONE);	// Accumulate all pixels togethers
  glBlendEquation(GL_FUNC_ADD);

  // Save the current viewport.
  int viewport[4];
  glGetIntegerv(GL_VIEWPORT,viewport);

  // Define the viewport to the region we are interested in.
  glViewport(0,0,textureSize_,textureSize_);

  // Set the current texture with the baseTexture
  glEnable(GL_TEXTURE_2D);
  glBindTexture(GL_TEXTURE_2D, paperTextureID);

  // Set the modulate blending method.
  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

  // Set the current color to a light gray.
  // this is the paper's diffuse color without texture.
  glColor4f(0.8f, 0.8f, 0.8f, 1.0f);

  // Save the modelview matrix and set it to identity.
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity();

  // Save the projecion matrix and set it to orthographic.
  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  glLoadIdentity();
  glOrtho(0.0f, 1.0f, 0.0f, 1.0f, -1.0f, 1.0f);

  // Save the texture matrix and set it to identity.
  glMatrixMode(GL_TEXTURE);
  glPushMatrix();
  glLoadIdentity();
  
  // Scaling factor between texture zoom / octave.
  float Fi = 1.0f;

  // Normalization factor so that Ai*(1 + 1/2 + ... + 1/2^(nbOctaves()-1)) == 1.0f
  float Ai= static_cast<float>(1<<nbOctaves()) / ( (1<<nbOctaves()) - 1.0 );

  // Set index to the first octave.
  unsigned int index = omega0;

  // Fi and Ai are updated according to section 5.2.
  for (unsigned int i=0; i<nbOctaves(); ++i)
    {
      // Make a cross-fade for the first and last octaves.
      // Other octaves are at 100%.
      const float alpha = (i==0) ? (zoom-1.0f)
	: (i==(nbOctaves()-1)) ? (2.0f-zoom)
	: 1.0f;

      glBlendColor(1.0f, 1.0f, 1.0f, alpha * Ai / zoom );
      
      // Randomize translation between octaves.
      const float u = randomOffsetX[index];
      const float v = randomOffsetY[index];
      // const float u = 0;
      // const float v = 0;

      // Draw the octave.
      glBegin(GL_QUADS);
      glTexCoord2f(u,    v);    glVertex3f(0.0f, 0.0f, 0.0f);
      glTexCoord2f(u+Fi, v);    glVertex3f(1.0f, 0.0f, 0.0f);
      glTexCoord2f(u+Fi, v+Fi); glVertex3f(1.0f, 1.0f, 0.0f);
      glTexCoord2f(u,    v+Fi); glVertex3f(0.0f, 1.0f, 0.0f);
      // glTexCoord2f(u,    v);    glVertex3f(i/3.0, 		0.0f, 0.0f);
      // glTexCoord2f(u+Fi, v);    glVertex3f(i/3.0+1/3.0, 	0.0f, 0.0f);
      // glTexCoord2f(u+Fi, v+Fi); glVertex3f(i/3.0+1/3.0, 	1.0/3.0, 0.0f);
      // glTexCoord2f(u,    v+Fi); glVertex3f(i/3.0, 		1.0/3.0, 0.0f);
      glEnd();

      // If we are working with the first octave we load it
      // else we accumulate.
      // glAccum( (i==0) ? GL_LOAD : GL_ACCUM, alpha * Ai / zoom);

      // Switch octave in a cyclic way.
      index = (index+1) % nbOffset;

      Fi *= 2.0;
      Ai *= 0.5;
    }

  // Dump the accumulation buffer values into the current buffer.
  // glAccum(GL_RETURN, 1.0f);

  // Flush the OpenGL execution queue.
  glFlush();

  // Set the texture engine with the multi-scale texture.
  glBindTexture(GL_TEXTURE_2D, msTextureID);

  // Copy the composed texture in the multi-scale texture.
  glCopyTexSubImage2D(GL_TEXTURE_2D,
		      0,	// Mipmap level.
		      0, 0,	// Offset.
		      0, 0,	// Origin
		      textureSize_, textureSize_);

  // Restore the matrices.
  glMatrixMode(GL_TEXTURE);    glPopMatrix();  
  glMatrixMode(GL_PROJECTION); glPopMatrix();	
  glMatrixMode(GL_MODELVIEW);  glPopMatrix();	

  // Restore the viewport,
  glViewport(viewport[0],viewport[1],viewport[2],viewport[3]);

  glDisable(GL_BLEND);
  glDisable(GL_TEXTURE_2D);

  glEnable(GL_DEPTH_TEST);
  glEnable(GL_COLOR_MATERIAL);
  glEnable(GL_LIGHTING);

  glClear(GL_COLOR_BUFFER_BIT);
}


void DynamicCanvas::setTextureMatrix()
{
  // Switch to GL_TEXTURE mode and initialize with identity matrix.
  glMatrixMode(GL_TEXTURE);
  glLoadIdentity();

  const float z = zoom/frequencyBase();
  const float R = z / tan(fieldOfView/2.0);
  const float d = R / tan(alpha);
  
  float localdX, localdY;
  if (fabs(alpha) < 1E-3)
    {
      // For small alpha values, a first order approximation is used5 * 
      // to prevent division by zero (see appendix A).
      localdX = R*deltaBeta;
      localdY = R*deltaAlpha;
    }
  else
    {
      localdX = d * sinf(deltaGamma);
      localdY = d * (1.0f - cosf(deltaGamma)) + R*deltaAlpha;
    }

  // Apply R_gamma rotation (Appendix A)
  float DX = cosf(gamma) * localdX - sinf(gamma) * localdY;
  float DY = sinf(gamma) * localdX + cosf(gamma) * localdY;
  
  // Update rectangle position.  
  posX += DX;
  posY += DY;

  // Update the matrix.
  glTranslatef(posX, posY, 0.0f);
  glRotatef(gamma * (180.0f/M_PI), 0, 0, 1);
  glScalef(z, z, 1.0f);
}


void DynamicCanvas::warpedTextureCoordinate(const float x, const float y)
{
  // Let's call M the point on the screen.
  // Compensate for the screen aspect ratio.
  const vectorType M(ratio*x, y, 0.0);

  // Center of the sphere relatively to the texture plane
  // on which the ball is rolling.
  const vectorType sphereCenter(0, 0, sphereRadius);

  // Compute the vector from the sphere center to M.
  // Remember: I is the short name of 'sphereCenter'.
  const vectorType IM = M - sphereCenter;

  // Let's call M' the projection of M on the sphere. M' is noted Mp.
  // IM' = (IM * sphereRadius) / ||IM||
  const vectorType IMp = (IM*sphereRadius)/IM.norm();


  // Vector pointing toward the north pole of the sphere.
  const vectorType upVector(0,cos(alpha),-sin(alpha));

  // Let's call "M bar" the projection of M' on the up vector.
  // IMbar = (IM'.upVector) * upVector
  const vectorType IMbar = (IMp*upVector) * upVector;

  // Compute the vector from the sphere center to the screen center (0,0,0).
  // Remember: O is the short name of 'screenCenter'.
  const vectorType IO = vectorType(0.0,0.0,0.0) - sphereCenter; 

  // Let's call "O bar" the projection of O on the up vector.
  // IMbar = (IO.upVector) * upVector
  const vectorType IObar = (IO*upVector) * upVector;

  // Compute (phi - alpha). See Appendix B in the paper.
  // sin(phi) = IMbar.upVector / sphereRadius
  const float phiMinusAlpha = asin((IMbar*upVector) / sphereRadius) - alpha;

  // Compute o. See Appendix B in the paper.
  const float o = phiMinusAlpha * sphereRadius;

  // Compute the vector from Mbar to M'.
  vectorType MbarMp = IMp-IMbar;

  // Normalize MbarM'. Avoid division by 0.
  if (MbarMp.norm()<1.E-10) 
    MbarMp = vectorType(0.0,0.0,0.0);
  else
    MbarMp.normalize();

    
  // Compute the vector from Obar to O.    
  // Remember: O is the short name of 'screenCenter'.
  vectorType ObarO = IO-IObar;

  // Normalize ObarO. Avoid division by 0.
  if (ObarO.norm()<1.E-10)
    ObarO = vectorType(0.0,0.0,0.0);
  else
    ObarO.normalize();

  // Compute theta (Appendix B) through sin and cos.
  // cos(theta) = MbarM'.ObarO
  const float cosTheta = MbarMp*ObarO;

  // sin(theta) = (MbarM' x ObarO) . upVector
  const float sinTheta = (MbarMp^ObarO) * upVector;

  // Classical atan2 from sin and cos.
  const float theta = atan2f(sinTheta,cosTheta);


  // Compute the point C (Appendix B).
  // C.x = C.z = 0
  // tan(alpha) = sphereRadius / C.y
  const vectorType C(0,(sphereRadius/tan(alpha)),0);

  // Compute d - o (Appendix B).
  const float dMinusO =  C.y - o;

  // Compute sigma (Appendix B).
  // Considering the screen center:
  // C.y * sigma = sphereRadius * cos(alpha)* theta
  const float sigma = theta * sphereRadius * cos(alpha)/C.y;

  // Compute u and v (Appendix B).
  // Because of the blending near the poles, we name them "warped".
  const float warpedU = dMinusO * sin(sigma);
  const float warpedV = o + dMinusO * (1 - cos(sigma));


    
  /////////
  // Compute the blending factor.
  /////////
    
  float blendFactor = 0.0f; // Default = full warp.

  // Angular deviation to the pole = PI/2 - |alpha|
  const float poleDeviation = M_PI_2 - fabsf(alpha);

  if (poleDeviation < 0.5*fieldOfView) // Pole is visible
    blendFactor = 1.0f; // No warp
  else
    if (poleDeviation < fieldOfView) // Near the pole
      blendFactor = (fieldOfView-poleDeviation) / (0.5*fieldOfView); // Blending

  // Send texture coordinates to OpenGL (linear blending).
  glTexCoord2f(blendFactor*M.x + (1.0-blendFactor)*warpedU,
	       blendFactor*M.y + (1.0-blendFactor)*warpedV);
}


void DynamicCanvas::unwarpedTextureCoordinate(const float x, const float y)
{  
  // Compensate for the screen aspect ratio.
  // Send texture coordinates to OpenGL (linear blending).
  glTexCoord2f(ratio*x, y);
}


void DynamicCanvas::textureCoordinate(const float x, const float y)
{
  if (warped)
    warpedTextureCoordinate(x,y);
  else
    unwarpedTextureCoordinate(x,y);
}
